from numpy.testing import assert_equal
from falass import dataformat
import unittest

class TestQData(unittest.TestCase):
    def test_qdata(self):
        a = dataformat.QData(1., 2., 3., 4.)
        assert_equal(a.q, 1.)
        assert_equal(a.i, 2.)
        assert_equal(a.di, 3.)
        assert_equal(a.dq, 4.)

class TestScatLens(unittest.TestCase):
    def test_scatlens(self):
        a = dataformat.ScatLens('C1', 1., 0.)
        assert_equal(a.atom, 'C1')
        assert_equal(a.real, 1.0e-5)
        assert_equal(a.imag, 0.0)

class TestAtomPositions(unittest.TestCase):
    def test_atompositions(self):
        a = dataformat.AtomPositions('C1', 1.000)
        assert_equal(a.atom, 'C1')
        assert_equal(a.zpos, 1.000)

class TestSLDPro(unittest.TestCase):
    def test_sldpro(self):
        a = dataformat.SLDPro(1., 2., 3.)
        assert_equal(a.thick, 1.)
        assert_equal(a.real, 2.)
        assert_equal(a.imag, 3.)


from numpy.testing import assert_almost_equal
from falass import dataformat, reflect
import unittest


class TestReflect(unittest.TestCase):
    def test_reflect(self):
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        data1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        data = [data1, data2, data3]
        a = reflect.Reflect(sld, data)
        assert_almost_equal(a.sld_profile[0][0].thick, 1.)
        assert_almost_equal(a.sld_profile[0][0].real, 0.)
        assert_almost_equal(a.sld_profile[0][0].imag, 0.)
        assert_almost_equal(a.sld_profile[0][1].thick, 1.)
        assert_almost_equal(a.sld_profile[0][1].real, 5.)
        assert_almost_equal(a.sld_profile[0][1].imag, 0.)
        assert_almost_equal(a.exp_data[0].q, 0.05)
        assert_almost_equal(a.exp_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.exp_data[1].q, 0.25)
        assert_almost_equal(a.exp_data[1].dq, 0.25 * 0.05)
        assert_almost_equal(a.exp_data[2].q, 0.50)
        assert_almost_equal(a.exp_data[2].dq, 0.50 * 0.05)

    def test_average_reflect(self):
        data11 = dataformat.QData(0.05, 0.1, 0., 0.05 * 0.05)
        data12 = dataformat.QData(0.25, 0.05, 0., 0.05 * 0.25)
        data13 = dataformat.QData(0.50, 0.01, 0., 0.05 * 0.50)
        data1 = [data11, data12, data13]
        data21 = dataformat.QData(0.05, 0.12, 0., 0.05 * 0.05)
        data22 = dataformat.QData(0.25, 0.03, 0., 0.05 * 0.25)
        data23 = dataformat.QData(0.50, 0.015, 0., 0.05 * 0.50)
        data2 = [data21, data22, data23]
        data31 = dataformat.QData(0.05, 0.14, 0., 0.05 * 0.05)
        data32 = dataformat.QData(0.25, 0.07, 0., 0.05 * 0.25)
        data33 = dataformat.QData(0.50, 0.005, 0., 0.05 * 0.50)
        data3 = [data31, data32, data33]
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        ddata2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        ddata3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        ddata = [ddata1, ddata2, ddata3]
        a = reflect.Reflect(sld, ddata)
        a.reflect = [data1, data2, data3]
        a.average_ref()
        assert_almost_equal(a.averagereflect[0].i, 0.12)
        assert_almost_equal(a.averagereflect[1].i, 0.05)
        assert_almost_equal(a.averagereflect[2].i, 0.01)




import numpy as np
from falass import readwrite, dataformat
import os


class Job():
    """The catch all.

    This class is used for setting up the falass job -- and is generally a catch all for inputs that do not fit into
    other parts.

    Parameters
    ----------
    files: falass.readwrite.Files
        A Files class item.
    layer_thickness: float
        The thickness of the layers that the simulation cell should be sliced into.
    cut_off_size: float
        The size of the simulation cell that should be ignored from the bottom -- this is to allow for the use
        of a vacuum gap at the bottom of the cell.
    """
    def __init__(self, files, layer_thickness, cut_off_size):
        self.files = files
        self.layer_thickness = layer_thickness
        self.cut_off_size = cut_off_size
        self.times = np.asarray(self.files.times)
        self.new_file = False

    def set_run(self, files=None, layer_thickness=None, cut_off_size=None):
        """Edit job inputs.

        This allows parts of the class to be assigned after the initial assignment or changed

        Parameters
        ----------
        files: falass.readwrite.Files
            A Files class item.
        layer_thickness: float
            The thickness of the layers that the simulation cell should be sliced into.
        cut_off_size: float
            The size of the simulation cell that should be ignored from the bottom -- this is to allow for the use
            of a vacuum gap at the bottom of the cell.
        """
        if files:
            self.files = files
            self.times = np.asarray(self.files.times)
        if layer_thickness:
            self.layer_thickness = layer_thickness
        if cut_off_size:
            self.cut_off_size = cut_off_size

    def set_lgts(self):
        """Assign scattering lengths.

        Assigned the scattering lengths from the lgtfile to the different atom types. If no lgtfile is defined falass
        will help the user to build one by working through the atom types in the pdb file and requesting input of the
        real and imaginary scattering lengths. This will also occur if a atom type if found in the pdbfile but not in
        the given lgts file. falass will write the lgtfile to disk if atom types do not feature in the given lgtfile or
        one is written from scratch.
        """
        if self.files.lgtfile:
            path, extension = os.path.splitext(self.files.lgtfile)
            lgtfile_name = path + extension
            for i in range(0, len(self.files.atoms)):
                for j in range(0, len(self.files.atoms[i])):
                    duplicate = readwrite.check_duplicates(self.files.scat_lens, self.files.atoms[i][j].atom)
                    if not duplicate:
                        self.new_file = True
                        real_scat_len = input('The following atom type has no scattering length given '
                                              'in the lgt file {} \nPlease define a real scattering length for '
                                              'this atom type: '.format(self.files.atoms[i][j].atom))
                        imag_scat_len = input('\nPlease define a imaginary scattering length for '
                                              'this atom type: '.format(self.files.atoms[i][j].atom))
                        self.files.scat_lens.append(dataformat.ScatLens(self.files.atoms[i][j].atom, float(real_scat_len),
                                                                  float(imag_scat_len)))
        else:
            self.new_file = True
            print('There was no lgt file defined, falass will help you define one and save it for future use.')
            for i in range(0, len(self.files.atoms)):
                for j in range(0, len(self.files.atoms[i])):
                    duplicate = check_duplicates(self.files.scat_lens, self.files.atoms[i][j].atom)
                    if not duplicate:
                        real_scat_len = input('The following atom type has no scattering length given '
                                              'in the lgt file {} \nPlease define a real scattering length for '
                                              'this atom type: '.format(self.files.atoms[i][j].atom))
                        imag_scat_len = input('\nPlease define a imaginary scattering length for '
                                              'this atom type: '.format(self.files.atoms[i][j].atom))
                        self.files.scat_lens.append(dataformat.ScatLens(self.files.atoms[i][j].atom, float(real_scat_len),
                                                                  float(imag_scat_len)))
            lgtfile_name = input("What should the lgt file be named? ")
            path, extension = os.path.splitext(lgtfile_name)
            if extension != '.lgt':
                lgtfile_name = path + '.lgt'
        if self.new_file:
            i = 0
            while os.path.isfile(lgtfile_name):
                i+=1
                lgtfile_name = path + str(i) + '.lgt'

            lgtsf = open(lgtfile_name, 'w')
            for i in range(0, len(self.files.scat_lens)):
                lgtsf.write('{} {} {}\n'.format(self.files.scat_lens[i].atom, self.files.scat_lens[i].real * 1e5,
                                                self.files.scat_lens[i].imag * 1e5))
            print('A new lgtfile has been written with the name {}'.format(lgtfile_name))

    def set_times(self, times=None):
        """Assign times to analyse.

        The assignment of the simulation timesteps that should be analysed. if none are given all will be analysed.

        Parameters
        ----------
        times: array_like float
            The timesteps that should be analysed, in the unit of time that present in the pdbfile.
        """
        if times:
            self.times = np.arange(float(times[0]), float(times[1]) + float(times[2]), float(times[2]))
        else:
            first_times = float(input(
                "Please define the first timestep to be analysed, the first in the pdb file was {} ps: ".format(
                    self.files.times[0])))
            while check_array(self.files.times, first_times) is not True:
                first_times = float(input("TIMESTEP NOT FOUND. Please define the first timestep to be analysed, "
                                    "the first in the pdb file was {} ps: ".format(self.files.times[0])))
            last_times = float(input(
                "Please define the last timestep to be analysed, the last in the pdb file was {} ps: ".format(
                    self.files.times[-1])))
            while check_array(self.files.times, last_times) is not True:
                last_times = float(input("TIMESTEP NOT FOUND. Please define the last timestep to be analysed, "
                                   "the last in the pdb file was {} ps: ".format(self.files.times[-1])))
            interval_times = float(input("Please define time interval for analysis, the smallest interval in the pdb "
                                   "file was {} ps: ".format(self.files.times[1] - self.files.times[0])))
            while interval_times > last_times:
                interval_times = float(input("NOT A VALID INTERVAL. Please define time interval for analysis, "
                                       "the smallest interval in the pdb file was {} ps: ".format(self.files.times[1] -
                                                                                                  self.files.times[0])))
            self.times = np.arange(first_times, last_times + interval_times, interval_times)


def check_array(array, check):
    """Checks if item is in array.

    Checks if an item has already been added to an array.

    Parameters
    ----------
    array: array-type
        The array to check.
    check: str
        The item to try and find.

    Returns
    -------
    bool
        true if the item is already present in the scatlen type array, false if not.
    """
    for i in range(0, len(array)):
        if check == array[i]:
            return True
    return False



class QData:
    """Reflectometry data.

    A class to hold the qdata information consisting of four floats associated with the q-vector, the intensity of the
    reflectometry, the uncertainty in the intensity and the uncertainty in the q-vector/resolution of the q-vector.
    """
    def __init__(self, q, i, di, dq):
        self.q = q
        self.i = i
        self.di = di
        self.dq = dq


class ScatLens:
    """Scattering lengths.

    A class to hold the scattering lengths of the different atom types consisting of a str atom type name, and two
    floats associated with the real and imaginary scattering lengths of that atom type.
    """
    def __init__(self, atom, real, imag):
        self.atom = atom
        self.real = real * 1e-5
        self.imag = imag * 1e-5


class AtomPositions:
    """z-Dimension positions.

    A class to hold the atom positions in the z-dimension, consisting of the str atom type name and a float giving the
    position in the z-dimension.
    """
    def __init__(self, atom, position):
        self.atom = atom
        self.zpos = position


class SLDPro:
    """Layer information for SLD profile.

    A class to hold the layer description of the sld profile consisting of three floats associated with the thickness,
    real scattering length density and imaginary scattering length density.
    """
    def __init__(self, thick, real, imag):
        self.thick = thick
        self.real = real
        self.imag = imag



from numpy.testing import assert_equal, assert_almost_equal
from falass import readwrite, job, sld, dataformat
import os
import unittest


class TestSLD(unittest.TestCase):
    def test_sld(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                            datfile=os.path.join(self.path, 'test3.dat'))
        a.read_pdb()
        a.read_lgt()
        a.read_dat()
        b = job.Job(a, 1., 5.)
        b.set_times(times = [0., 20000., 10000.])
        b.set_lgts()
        c = sld.SLD(b)
        assert_equal(len(c.assigned_job.files.times), 6)
        assert_equal(c.assigned_job.files.times, [0, 10000., 20000., 30000., 40000., 50000.])
        assert_equal(len(c.assigned_job.times), 3)
        assert_equal(c.assigned_job.times, [0, 10000., 20000.])
        assert_equal(c.assigned_job.layer_thickness, 1.)
        assert_equal(c.assigned_job.cut_off_size, 5.)

    def test_get_sld_profile(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                            datfile=os.path.join(self.path, 'test3.dat'))
        a.read_pdb()
        a.read_lgt()
        a.read_dat()
        b = job.Job(a, 1., 0.)
        b.set_times(times=[0., 20000., 10000.])
        b.set_lgts()
        c = sld.SLD(b)
        c.get_sld_profile()
        assert_equal(len(c.sld_profile), 3)
        for i in range(0, len(c.sld_profile)):
            assert_equal(len(c.sld_profile[i]), 4)
        assert_almost_equal(c.sld_profile[0][0].thick, 1.)
        assert_almost_equal(c.sld_profile[0][0].real, 0.)
        assert_almost_equal(c.sld_profile[0][0].imag, 0.)
        assert_almost_equal(c.sld_profile[0][1].thick, 1.)
        assert_almost_equal(c.sld_profile[0][1].real, 1e-5)
        assert_almost_equal(c.sld_profile[0][1].imag, 0.)
        assert_almost_equal(c.sld_profile[0][2].thick, 1.)
        assert_almost_equal(c.sld_profile[0][2].real, 2e-5)
        assert_almost_equal(c.sld_profile[0][2].imag, 1e-5)
        assert_almost_equal(c.sld_profile[0][3].thick, 1.)
        assert_almost_equal(c.sld_profile[0][3].real, 3e-5)
        assert_almost_equal(c.sld_profile[0][3].imag, 2e-5)
        assert_almost_equal(c.sld_profile[1][0].thick, 1.)
        assert_almost_equal(c.sld_profile[1][0].real, 0. / 2.)
        assert_almost_equal(c.sld_profile[1][0].imag, 0. / 2.)
        assert_almost_equal(c.sld_profile[1][1].thick, 1.)
        assert_almost_equal(c.sld_profile[1][1].real, 2e-5 / 2.)
        assert_almost_equal(c.sld_profile[1][1].imag, 1e-5 / 2.)
        assert_almost_equal(c.sld_profile[1][2].thick, 1.)
        assert_almost_equal(c.sld_profile[1][2].real, 3e-5 / 2.)
        assert_almost_equal(c.sld_profile[1][2].imag, 2e-5 / 2.)
        assert_almost_equal(c.sld_profile[1][3].thick, 1.)
        assert_almost_equal(c.sld_profile[1][3].real, 1e-5 / 2.)
        assert_almost_equal(c.sld_profile[1][3].imag, 0 / 2.)
        assert_almost_equal(c.sld_profile[2][0].thick, 1.)
        assert_almost_equal(c.sld_profile[2][0].real, 0. / 3.)
        assert_almost_equal(c.sld_profile[2][0].imag, 0. / 3.)
        assert_almost_equal(c.sld_profile[2][1].thick, 1.)
        assert_almost_equal(c.sld_profile[2][1].real, 3e-5 / 3.)
        assert_almost_equal(c.sld_profile[2][1].imag, 2e-5 / 3.)
        assert_almost_equal(c.sld_profile[2][2].thick, 1.)
        assert_almost_equal(c.sld_profile[2][2].real, 1e-5 / 3.)
        assert_almost_equal(c.sld_profile[2][2].imag, 0 / 3.)
        assert_almost_equal(c.sld_profile[2][3].thick, 1.)
        assert_almost_equal(c.sld_profile[2][3].real, 2e-5 / 3.)
        assert_almost_equal(c.sld_profile[2][3].imag, 1e-5 / 3.)

    def test_average_sld_profile(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                            datfile=os.path.join(self.path, 'test3.dat'))
        a.read_pdb()
        a.read_lgt()
        a.read_dat()
        b = job.Job(a, 1., 0.)
        b.set_times(times=[0., 20000., 10000.])
        b.set_lgts()
        c = sld.SLD(b)
        c.get_sld_profile()
        c.average_sld_profile()
        assert_equal(len(c.av_sld_profile), 4)
        assert_almost_equal(c.av_sld_profile[0].real, 0.)
        assert_almost_equal(c.av_sld_profile[0].imag, 0.)
        assert_almost_equal(c.av_sld_profile[1].real, (1e-5 + (2e-5 / 2.) + (3e-5 / 3.)) / 3.)
        assert_almost_equal(c.av_sld_profile[1].imag, (0 + (1e-5 / 2.) + (2e-5 / 3.)) / 3.)
        assert_almost_equal(c.av_sld_profile[2].real, (2e-5 + (3e-5 / 2.) + (1e-5 / 3.)) / 3.)
        assert_almost_equal(c.av_sld_profile[2].imag, ((1e-5) + (2e-5 / 2.) + (0 / 3.)) / 3.)
        assert_almost_equal(c.av_sld_profile[3].real, (3e-5 + (1e-5 / 2.) + (2e-5 / 3.)) / 3.)
        assert_almost_equal(c.av_sld_profile[3].imag, ((2e-5) + (0 / 2.) + (1e-5 / 3.)) / 3.)


def test_get_scatlen():
    atom1 = dataformat.ScatLens('C1', 1.0, 0.0)
    atom2 = dataformat.ScatLens('C2', 2.0, 1.0)
    atom3 = dataformat.ScatLens('C3', 3.0, 2.0)
    array = [atom1, atom2, atom3]
    real, imag = sld.get_scatlen('C3', array)
    assert_almost_equal(real, 3.0e-5)
    assert_almost_equal(imag, 2.0e-5)


from numpy.testing import assert_almost_equal
from falass import compare, dataformat
import numpy as np
import unittest

class TestCompare(unittest.TestCase):
    def test_compare(self):
        data1 = dataformat.QData(0.05, 3., 0.3, 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 1., 0.1, 0.05 * 0.50)
        data = [data1, data2, data3]
        sdata1 = dataformat.QData(0.05, 1., 0.1, 0.05 * 0.05)
        sdata2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        sdata3 = dataformat.QData(0.50, 3., 0.3, 0.05 * 0.50)
        sdata = [sdata1, sdata2, sdata3]
        a = compare.Compare(data, sdata, 1., 0.)
        assert_almost_equal(a.exp_data[0].q, 0.05)
        assert_almost_equal(a.exp_data[0].i, 3.)
        assert_almost_equal(a.exp_data[0].di, 0.3)
        assert_almost_equal(a.exp_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.exp_data[1].q, 0.25)
        assert_almost_equal(a.exp_data[1].i, 2.)
        assert_almost_equal(a.exp_data[1].di, 0.2)
        assert_almost_equal(a.exp_data[1].dq, 0.05 * 0.25)
        assert_almost_equal(a.exp_data[2].q, 0.50)
        assert_almost_equal(a.exp_data[2].i, 1.)
        assert_almost_equal(a.exp_data[2].di, 0.1)
        assert_almost_equal(a.exp_data[2].dq, 0.05 * 0.50)
        assert_almost_equal(a.sim_data[0].q, 0.05)
        assert_almost_equal(a.sim_data[0].i, 1.)
        assert_almost_equal(a.sim_data[0].di, 0.1)
        assert_almost_equal(a.sim_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.sim_data[1].q, 0.25)
        assert_almost_equal(a.sim_data[1].i, 2.)
        assert_almost_equal(a.sim_data[1].di, 0.2)
        assert_almost_equal(a.sim_data[1].dq, 0.05 * 0.25)
        assert_almost_equal(a.sim_data[2].q, 0.50)
        assert_almost_equal(a.sim_data[2].i, 3.)
        assert_almost_equal(a.sim_data[2].di, 0.3)
        assert_almost_equal(a.sim_data[2].dq, 0.05 * 0.50)
        assert_almost_equal(a.scale, 1.)
        assert_almost_equal(a.background, 0.)

    def test_change_scale(self):
        data1 = dataformat.QData(0.05, 3., 0.3, 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 1., 0.1, 0.05 * 0.50)
        data = [data1, data2, data3]
        sdata1 = dataformat.QData(0.05, 1., 0.1, 0.05 * 0.05)
        sdata2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        sdata3 = dataformat.QData(0.50, 3., 0.3, 0.05 * 0.50)
        sdata = [sdata1, sdata2, sdata3]
        a = compare.Compare(data, sdata, 1., 0.)
        assert_almost_equal(a.scale, 1.)
        a.change_scale(2.)
        assert_almost_equal(a.scale, 2.)

    def test_change_background(self):
        data1 = dataformat.QData(0.05, 3., 0.3, 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 1., 0.1, 0.05 * 0.50)
        data = [data1, data2, data3]
        sdata1 = dataformat.QData(0.05, 1., 0.1, 0.05 * 0.05)
        sdata2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        sdata3 = dataformat.QData(0.50, 3., 0.3, 0.05 * 0.50)
        sdata = [sdata1, sdata2, sdata3]
        a = compare.Compare(data, sdata, 1., 0.)
        assert_almost_equal(a.background, 0.)
        a.change_background(2.)
        assert_almost_equal(a.background, 2.)

def test_scale_and_background():
    a = np.array([1., 2., 3.])
    scale = 2.
    background = 1.
    b = compare.scale_and_background(a, scale, background)
    assert_almost_equal(b[0], 1.098612289)
    assert_almost_equal(b[1], 1.609437912)
    assert_almost_equal(b[2], 1.945910149)



import numpy as np
from falass import dataformat
import matplotlib.pyplot as plt
from scipy.interpolate import InterpolatedUnivariateSpline


class Reflect:
    """Reflectometry calculation.

    A class for the calculation of reflectometry from the sld profile defined in the falass.sld.SLD class.

    Parameters
    ----------
    sld_profile: array_like falass.dataformat.SLDPro
        An array describing the scattering length density of the simulation cell under study.
    exp_data: array_like falass.dataformat.QData
        An array giving the experimental data from the datfile.
    """
    def __init__(self, sld_profile, exp_data):
        self.sld_profile = sld_profile
        self.exp_data = exp_data
        self.averagereflect = []
        self.reflect = []

    def calc_ref(self):
        """Calculate reflectometry.

        The calculation of the reflectometry profiles based on the sld profiles calculated from each of the timesteps
        under study.
        """
        if len(self.exp_data) > 0:
            self.reflect = []
            prog = 0
            k = 0
            print("[ 0 % ]")
            for i in range(0, len(self.sld_profile)):
                k += 1
                prog_new = np.floor(k / (len(self.sld_profile)) * 100)
                if prog_new > prog + 9:
                    prog = prog_new
                    print("[{} {} % ]".format('#' * int(prog / 10), int(prog / 10) * 10))
                refl = convolution(self.exp_data, self.sld_profile[i])
                a = []
                for j in range(0, len(self.exp_data)):
                    a.append(dataformat.QData(self.exp_data[j].q, refl[j], 0, self.exp_data[j].dq))
                self.reflect.append(a)
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')

    def average_ref(self):
        """Average reflectometry profiles.

        The averaging of the reflectometry profiles as calculated by the calc_ref() function.
        """
        if len(self.exp_data) > 0:
            self.averagereflect = []
            for i in range(0, len(self.reflect[0])):
                self.averagereflect.append(dataformat.QData(self.exp_data[i].q, 0, 0, self.exp_data[i].dq))
            for i in range(0, len(self.reflect[0])):
                for j in range(0, len(self.reflect)):
                    self.averagereflect[i].i += self.reflect[j][i].i
                self.averagereflect[i].i /= len(self.reflect)
                for j in range(0, len(self.reflect)):
                    self.averagereflect[i].di += np.square(self.reflect[j][i].i - self.averagereflect[i].i)
                self.averagereflect[i].di = np.sqrt(1. / (len(self.reflect) - 1) * self.averagereflect[i].di)
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')

    def plot_ref(self, rq4=True):
        """Plot reflectometry profile.

        The plotting of the calculated reflectometry data without comparison to experimental data.

        Parameters
        ----------
        rq4: bool
            Should the data be transformed to rq4 space.
        """
        if len(self.exp_data) > 0:
            x = []
            y = []
            dy = []
            plt.rc('text', usetex=True)
            plt.rc('font', family='serif')
            if rq4:
                for i in range(0, len(self.exp_data)):
                    x.append(self.averagereflect[i].q)
                    y.append(np.log10(self.averagereflect[i].i * self.averagereflect[i].q ** 4))
                    dy.append((self.averagereflect[i].di * self.averagereflect[i].q ** 4) /
                              (self.averagereflect[i].i * np.log(10)))
            else:
                for i in range(0, len(self.exp_data)):
                    x.append(self.averagereflect[i].q)
                    y.append(np.log10(self.averagereflect[i].i))
                    dy.append(self.averagereflect[i].di / (self.averagereflect[i].i * np.log(10)))
            x = np.asarray(x)
            y = np.asarray(y)
            dy = np.asarray(dy)
            plt.errorbar(x, y, yerr=dy)
            plt.xlabel('$q$ (\AA)')
            plt.ylabel('log($Rq^4$) (\AA$^4$)')
            plt.show()
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')


def convolution(exp_data, sld_profile):
    """Convolution/smearing

    The convolution of the reflectometry data by a gaussian of constant width (a percentage of the q-vector)

    Parameters
    ----------
    exp_data: falass.dataformat.QData
        The experimental data from the datfile.
    sld_profile: falass.dataformat.SLDPro
        The SLD profile calculated from the simulation trajectory.
    lt: float
        The thickness of the layers.

    Returns
    -------
    array_like
        The smeared reflectometry profile.
    """
    fwhm = 2 * np.sqrt(2 * np.log(2))

    res = exp_data[0].dq / exp_data[0].q

    if exp_data[0].dq / exp_data[0].q < 0.0005:
        return reflectivity(exp_data, sld_profile)

    gnum = 51
    ggpoint = (gnum - 1) / 2

    def gauss(x, s):
        return 1. / s / np.sqrt(2 * np.pi) * np.exp(-0.5 * x ** 2 / s / s)

    q = []
    for i in range(0, len(exp_data)):
        q.append(exp_data[i].q)

    lowq = np.min(q)
    highq = np.max(q)

    start = np.log10(lowq) - 6 * res / fwhm
    finish = np.log10(highq * (1 + 6 * res / fwhm))
    interp = np.round(np.abs(1 * (np.abs(start-finish)) / (1.7 * res / fwhm / ggpoint)))

    xtemp = np.linspace(start, finish, int(interp))
    xlin = np.power(10., xtemp)

    gaussx = np.linspace(-1.7 * res, 1.7 * res, gnum)
    gaussy = gauss(gaussx, res / fwhm)

    rvals = reflectivity(xlin, sld_profile)
    smeared_rvals = np.convolve(rvals, gaussy, mode='same')
    interpol = InterpolatedUnivariateSpline(xlin, smeared_rvals)

    smeared_output = interpol(q)
    smeared_output *= gaussx[1] - gaussx[0]
    return smeared_output


def reflectivity(exp_data, sld_profile):
    """Abeles optical matrix formalism.

    The calculation of the reflectometry using the Abeles optical matrix method.

    Parameters
    ----------
    exp_data: falass.dataformat.QData
        The experimental data from the datfile.
    sld_profile: falass.dataformat.SLDPro
        The SLD profile calculated from the simulation trajectory.
    lt: float
        The thickness of the layers.

    Returns
    -------
    array_like
        The reflectometry profile.
    """
    layers = np.zeros((len(sld_profile), 4))
    for i in range(0, len(sld_profile)):
        layers[i][0] = sld_profile[i].thick
        layers[i][1] = sld_profile[i].real
        layers[i][2] = sld_profile[i].imag
        layers[i][3] = 0
    qvals = np.asfarray(exp_data).ravel()
    nlayers = len(sld_profile) - 2
    npnts = qvals.size

    kn = np.zeros((npnts, nlayers + 2), np.complex128)

    sld = np.zeros(nlayers + 2, np.complex128)
    sld[:] += ((layers[:, 1].real - layers[0, 1]) + 1j * (layers[:, 2] - layers[0, 2]))

    kn[:] = np.sqrt(qvals[:, np.newaxis] ** 2. / 4. - 4 * np.pi * sld)

    mrtot00 = 1
    mrtot11 = 1
    mrtot10 = 0
    mrtot01 = 0
    k = kn[:, 0]

    for idx in range(1, nlayers + 2):
        k_next = kn[:, idx]
        rj = (k - k_next) / (k + k_next)
        rj *= np.exp(k * k_next)

        # work out characteristic matrix of layer
        mi00 = np.exp(k * 1j * np.fabs(layers[idx - 1, 0])) if idx - 1 else 1
        mi11 = np.exp(k * -1j * np.fabs(layers[idx - 1, 0])) if idx - 1 else 1

        mi10 = rj * mi00
        mi01 = rj * mi11

        # matrix multiply mrtot by characteristic matrix
        p0 = mrtot00 * mi00 + mrtot10 * mi01
        p1 = mrtot00 * mi10 + mrtot10 * mi11
        mrtot00 = p0
        mrtot10 = p1

        p0 = mrtot01 * mi00 + mrtot11 * mi01
        p1 = mrtot01 * mi10 + mrtot11 * mi11

        mrtot01 = p0
        mrtot11 = p1

        k = k_next

    ref = (mrtot01 * np.conj(mrtot01)) / (mrtot00 * np.conj(mrtot00))
    return np.real(np.reshape(ref, exp_data.shape))



import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from falass import dataformat


class Compare:
    """Comparison and fitting.

    For the comparison and fitting of calculated and experimental reflectometry.

    Parameters
    ----------
    exp_data: array_like falass.dataformat.QData
        The experimental reflectometry data read from the datfile.
    sim_data: array_like falass.dataformat.QData
        The calculated reflectometry data from the simulation.
    scale: float
        The amount by which the calculated reflectometry should be scaled.
    background: float
        The height of the uniform background to be added to the calculated reflectometry.
    """
    def __init__(self, exp_data, sim_data, scale, background):
        self.exp_data = exp_data
        self.sim_data = sim_data
        self.scale = scale
        self.background = background
        self.sim_data_fitted = []

    def change_scale(self, scale):
        """Edit scale.

        Lets the scale factor be changed.

        Parameters
        ----------
        scale: float
            The amount by which the calculated reflectometry should be scaled.
        """
        self.scale = scale

    def change_background(self, background):
        """Edit background.

        Lets the background factor be changed.

        Parameters
        ----------
        background: float
            The height of the uniform background to be added to the calculated reflectometry.
        """
        self.background = background

    def fit(self):
        """Fit scale and background.

        Perform the fitting of the scale and background for the calculated data to the experimental data.
        Currently only a logarithmically transformed fitted can be conducted.
        """
        if len(self.exp_data) > 0:
            if self.exp_data[0].i is not None:
                y = []
                dy = []
                y2 = []
                for i in range(0, len(self.exp_data)):
                    y.append(np.log(self.exp_data[i].i))
                    dy.append(self.exp_data[i].di / (self.exp_data[i].i * np.log(10)))
                    y2.append(self.sim_data[i].i)
                popt, pcov = curve_fit(scale_and_background, y2, y, bounds=((1e-6, 0), (10, 1e-3)), sigma=dy)
                self.scale = popt[0]
                self.background = popt[1]
            else:
                raise ValueError('No experimental data has been set for comparison, please read in a a .dat file.')
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')

    def plot_compare(self, rq4=True, fitted=True):
        """Plot a comparison.

        Plotting the comparision between the calculated and experimental reflectometry.

        Parameters
        ----------
        rq4: bool
            Should the data be plotted in rq4 space.
        fitted: bool
            Should the fitted reflectometry data be used.
        """
        x = []
        y = []
        dy = []
        x2 = []
        y2 = []
        dy2 = []
        plt.rc('text', usetex=True)
        plt.rc('font', family='serif')
        if fitted:
            if len(self.sim_data_fitted) > 0:
                k = self.sim_data_fitted
            else:
                raise ValueError("The reflectometry data has not been returned yet -- please run the fit() function "
                                 "and the return_fitted().")
        else:
            k = self.sim_data
        if rq4:
            for i in range(0, len(self.exp_data)):
                x.append(k[i].q)
                y.append(k[i].i * np.power(self.exp_data[i].q, 4))
                da = k[i].di * np.power(self.exp_data[i].q, 4)
                dy.append(da)
                x2.append(self.exp_data[i].q)
                y2.append(self.exp_data[i].i * np.power(self.exp_data[i].q, 4))
                da = self.exp_data[i].di * np.power(self.exp_data[i].q, 4)
                dy2.append(da)
                plt.ylabel('$Rq^4$')
        else:
            for i in range(0, len(self.exp_data)):
                x.append(k[i].q)
                y.append(k[i].i)
                da = k[i].di
                dy.append(da)
                x2.append(self.exp_data[i].q)
                y2.append(np.log10(self.exp_data[i].i))
                da = self.exp_data[i].di
                dy2.append(da)
                plt.ylabel('log$_{10}$($R$)')
        x = np.asarray(x)
        y = np.asarray(y)
        dy = np.asarray(dy)
        x2 = np.asarray(x2)
        y2 = np.asarray(y2)
        dy2 = np.asarray(dy2)
        plt.errorbar(x, y, yerr=dy)
        plt.errorbar(x2, y2, yerr=dy2, linestyle='', marker='o')
        plt.xlabel('$q$ (\AA)')
        plt.yscale('log')
        plt.show()

    def return_fitted(self):
        """Return fitted.

        Return the fitted calculated reflectometry data for use.
        """
        self.sim_data_fitted = []
        for i in range(0, len(self.sim_data)):
            a = self.sim_data[i].i * self.scale + self.background
            b = self.sim_data[i].di * self.scale
            self.sim_data_fitted.append(dataformat.QData(self.sim_data[i].q, a, b, self.sim_data[i].dq))


def scale_and_background(sim_data, scale, background):
    """Apply scale and background.

    Apply a scale factor and uniform background to the calculated reflectometry data.

    Parameters
    ----------
    sim_data: array_like float
        The data to be scaled and have a background added.
    scale: float
        The amount by which the data should be scaled.
    background: float
        The size of the uniform background to be added.

    Returns
    -------
    array_like float
        The scaled and background added reflectometry in log space.
    """
    sim_data = np.log(sim_data * scale + background)
    return sim_data



from numpy.testing import assert_equal, assert_almost_equal
from falass import readwrite, dataformat
import os
import unittest


class TestFiles(unittest.TestCase):
    def test_files_standard(self):
        pdb = readwrite.Files('test.pdb')
        assert_equal(pdb.pdbfile, 'test.pdb')
        assert_equal(pdb.lgtfile, None)
        assert_equal(pdb.datfile, None)
        assert_equal(pdb.resolution, 5.)
        assert_equal(pdb.ierror, 5.)
        assert_equal(pdb.flip, False)
        return

    def test_files_variation(self):
        pdb = readwrite.Files('test1.pdb', lgtfile='test1.lgt', datfile='test1.dat', resolution=1., ierror=1.,
                              flip=True)
        assert_equal(pdb.pdbfile, 'test1.pdb')
        assert_equal(pdb.lgtfile, 'test1.lgt')
        assert_equal(pdb.datfile, 'test1.dat')
        assert_equal(pdb.resolution, 1.)
        assert_equal(pdb.ierror, 1.)
        assert_equal(pdb.flip, True)
        return

    def test_set_pdbfile(self):
        pdb = readwrite.Files('test.pdb')
        assert_equal(pdb.pdbfile, 'test.pdb')
        pdb.set_file(pdbfile='test1.pdb')
        assert_equal(pdb.pdbfile, 'test1.pdb')
        return

    def test_set_lgtfile(self):
        pdb = readwrite.Files('test.pdb')
        assert_equal(pdb.lgtfile, None)
        pdb.set_file(lgtfile='test1.lgt')
        assert_equal(pdb.lgtfile, 'test1.lgt')
        return

    def test_set_datfile(self):
        pdb = readwrite.Files('test.pdb')
        assert_equal(pdb.datfile, None)
        pdb.set_file(datfile='test1.dat')
        assert_equal(pdb.datfile, 'test1.dat')
        return

    def test_read_pdb(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files(os.path.join(self.path, 'test.pdb'))
        pdb.read_pdb()
        assert_equal(pdb.number_of_timesteps, 6)
        assert_equal(pdb.times, [0., 10000., 20000., 30000., 40000., 50000.])
        assert_equal(len(pdb.atoms), 6)
        for i in range(0, len(pdb.atoms)):
            assert_equal(len(pdb.atoms[i]), 3)
            assert_equal(pdb.atoms[i][0].atom, 'C1')
            assert_equal(pdb.atoms[i][1].atom, 'C2')
            assert_equal(pdb.atoms[i][2].atom, 'C3')
        for i in range(0, 2):
            assert_equal(pdb.atoms[i * 3][0].zpos, 01.500)
            assert_equal(pdb.atoms[i * 3][1].zpos, 02.500)
            assert_equal(pdb.atoms[i * 3][2].zpos, 03.500)
            assert_equal(pdb.atoms[i * 3 + 1][0].zpos, 03.500)
            assert_equal(pdb.atoms[i * 3 + 1][1].zpos, 01.500)
            assert_equal(pdb.atoms[i * 3 + 1][2].zpos, 02.500)
            assert_equal(pdb.atoms[i * 3 + 2][0].zpos, 02.500)
            assert_equal(pdb.atoms[i * 3 + 2][1].zpos, 03.500)
            assert_equal(pdb.atoms[i * 3 + 2][2].zpos, 01.500)
        assert_equal(pdb.cell, [[1.000, 1.000, 4.000], [2.000, 1.000, 4.000], [3.000, 1.000, 4.000],
                                [4.000, 1.000, 4.000], [5.000, 1.000, 4.000], [6.000, 1.000, 4.000]])
        return

    def test_read_pdb_flip(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files(os.path.join(self.path, 'test.pdb'), flip=True)
        pdb.read_pdb()
        assert_equal(pdb.number_of_timesteps, 6)
        assert_equal(pdb.times, [0., 10000., 20000., 30000., 40000., 50000.])
        assert_equal(len(pdb.atoms), 6)
        for i in range(0, len(pdb.atoms)):
            assert_equal(len(pdb.atoms[i]), 3)
            assert_equal(pdb.atoms[i][0].atom, 'C1')
            assert_equal(pdb.atoms[i][1].atom, 'C2')
            assert_equal(pdb.atoms[i][2].atom, 'C3')
        for i in range(0, 2):
            assert_equal(pdb.atoms[i * 3][0].zpos, 02.500)
            assert_equal(pdb.atoms[i * 3][1].zpos, 01.500)
            assert_equal(pdb.atoms[i * 3][2].zpos, 00.500)
            assert_equal(pdb.atoms[i * 3 + 1][0].zpos, 00.500)
            assert_equal(pdb.atoms[i * 3 + 1][1].zpos, 02.500)
            assert_equal(pdb.atoms[i * 3 + 1][2].zpos, 01.500)
            assert_equal(pdb.atoms[i * 3 + 2][0].zpos, 01.500)
            assert_equal(pdb.atoms[i * 3 + 2][1].zpos, 00.500)
            assert_equal(pdb.atoms[i * 3 + 2][2].zpos, 02.500)
        assert_equal(pdb.cell, [[1.000, 1.000, 4.000], [2.000, 1.000, 4.000], [3.000, 1.000, 4.000],
                                [4.000, 1.000, 4.000], [5.000, 1.000, 4.000], [6.000, 1.000, 4.000]])
        return

    def test_read_lgt(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files('test.pdb', lgtfile=os.path.join(self.path, 'test.lgt'))
        pdb.read_lgt()
        assert_equal(pdb.scat_lens[0].atom, 'C1')
        assert_almost_equal(pdb.scat_lens[0].real, 1e-5)
        assert_almost_equal(pdb.scat_lens[0].imag, 0)
        assert_equal(pdb.scat_lens[1].atom, 'C2')
        assert_almost_equal(pdb.scat_lens[1].real, 2e-5)
        assert_almost_equal(pdb.scat_lens[1].imag, 1e-5)
        assert_equal(pdb.scat_lens[2].atom, 'C3')
        assert_almost_equal(pdb.scat_lens[2].real, 3e-5)
        assert_almost_equal(pdb.scat_lens[2].imag, 2e-5)
        return

    def test_read_lgt_not_defined(self):
        pdb = readwrite.Files('test.pdb')
        with self.assertRaises(ValueError) as context:
            pdb.read_lgt()
        self.assertTrue("No lgtfile has been defined" in str(context.exception))
        return

    def test_read_dat2(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files('test.pdb', datfile=os.path.join(self.path, 'test2.dat'))
        pdb.read_dat()
        assert_almost_equal(pdb.expdata[0].q, 0.051793)
        assert_almost_equal(pdb.expdata[0].i, 0.00034985)
        assert_almost_equal(pdb.expdata[0].di, 0.00034985 * 0.05)
        assert_almost_equal(pdb.expdata[0].dq, 0.051793 * 0.05)
        assert_almost_equal(pdb.expdata[1].q, 0.13742)
        assert_almost_equal(pdb.expdata[1].i, 1.4924e-5)
        assert_almost_equal(pdb.expdata[1].di, 1.4924e-5 * 0.05)
        assert_almost_equal(pdb.expdata[1].dq, 0.13742 * 0.05)
        assert_almost_equal(pdb.expdata[2].q, 0.38285)
        assert_almost_equal(pdb.expdata[2].i, 1.215e-6)
        assert_almost_equal(pdb.expdata[2].di, 1.215e-6 * 0.05)
        assert_almost_equal(pdb.expdata[2].dq, 0.38285 * 0.05)
        return

    def test_read_dat3(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files('test.pdb', datfile=os.path.join(self.path, 'test3.dat'))
        pdb.read_dat()
        assert_almost_equal(pdb.expdata[0].q, 0.051793)
        assert_almost_equal(pdb.expdata[0].i, 0.00034985)
        assert_almost_equal(pdb.expdata[0].di, 5.0354e-6)
        assert_almost_equal(pdb.expdata[0].dq, 0.051793 * 0.05)
        assert_almost_equal(pdb.expdata[1].q, 0.13742)
        assert_almost_equal(pdb.expdata[1].i, 1.4924e-5)
        assert_almost_equal(pdb.expdata[1].di, 3.7729e-7)
        assert_almost_equal(pdb.expdata[1].dq, 0.13742 * 0.05)
        assert_almost_equal(pdb.expdata[2].q, 0.38285)
        assert_almost_equal(pdb.expdata[2].i, 1.215e-6)
        assert_almost_equal(pdb.expdata[2].di, 3.478e-7)
        assert_almost_equal(pdb.expdata[2].dq, 0.38285 * 0.05)
        return

    def test_read_dat4(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        pdb = readwrite.Files('test.pdb', datfile=os.path.join(self.path, 'test4.dat'))
        pdb.read_dat()
        assert_almost_equal(pdb.expdata[0].q, 0.051793)
        assert_almost_equal(pdb.expdata[0].i, 0.00034985)
        assert_almost_equal(pdb.expdata[0].di, 5.0354e-6)
        assert_almost_equal(pdb.expdata[0].dq, 0.0003)
        assert_almost_equal(pdb.expdata[1].q, 0.13742)
        assert_almost_equal(pdb.expdata[1].i, 1.4924e-5)
        assert_almost_equal(pdb.expdata[1].di, 3.7729e-7)
        assert_almost_equal(pdb.expdata[1].dq, 0.0007)
        assert_almost_equal(pdb.expdata[2].q, 0.38285)
        assert_almost_equal(pdb.expdata[2].i, 1.215e-6)
        assert_almost_equal(pdb.expdata[2].di, 3.478e-7)
        assert_almost_equal(pdb.expdata[2].dq, 0.0019)
        return

    def test_read_dat_not_defined(self):
        pdb = readwrite.Files('test.pdb')
        assert_equal(pdb.expdata, [])
        return

    def test_get_qs(self):
        pdb = readwrite.Files('test.pdb')
        pdb.get_qs(0.0005, 0.0020, 4)
        assert_equal(len(pdb.expdata), 4)
        assert_equal(pdb.expdata[0].q, 0.0005)
        assert_almost_equal(pdb.expdata[0].dq, 0.0005 * 0.05)
        assert_equal(pdb.expdata[1].q, 0.0010)
        assert_almost_equal(pdb.expdata[1].dq, 0.0010 * 0.05)
        assert_equal(pdb.expdata[2].q, 0.0015)
        assert_almost_equal(pdb.expdata[2].dq, 0.0015 * 0.05)
        assert_equal(pdb.expdata[3].q, 0.0020)
        assert_almost_equal(pdb.expdata[3].dq, 0.0020 * 0.05)
        return

    if __name__ == "__main__":
        unittest.main()


def test_check_duplicates_true():
    atom1 = dataformat.ScatLens('C1', 1.0, 0.0)
    atom2 = dataformat.ScatLens('C2', 2.0, 1.0)
    atom3 = dataformat.ScatLens('C3', 3.0, 2.0)
    array = [atom1, atom2, atom3]
    check = dataformat.ScatLens('C2', 2.0, 1.0)
    bool_ret = readwrite.check_duplicates(array, check.atom)
    assert_equal(bool_ret, True)
    return


def test_check_duplicates_false():
    atom1 = dataformat.ScatLens('C1', 1.0, 0.0)
    atom2 = dataformat.ScatLens('C2', 2.0, 1.0)
    atom3 = dataformat.ScatLens('C3', 3.0, 2.0)
    array = [atom1, atom2, atom3]
    check = dataformat.ScatLens('C4', 4.0, 3.0)
    bool_ret = readwrite.check_duplicates(array, check.atom)
    assert_equal(bool_ret, False)
    return


def test_line_count():
    path = os.path.dirname(os.path.abspath(__file__))
    lines = readwrite.line_count(os.path.join(path, 'test.pdb'))
    assert_equal(lines, 60)
    return



from numpy.testing import assert_equal
from falass import readwrite, job
import os
import unittest

class TestJob(unittest.TestCase):
    def test_job(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                              datfile=os.path.join(self.path, 'test3.dat'))
        a.read_pdb()
        a.read_lgt()
        a.read_dat()
        b = job.Job(a, 1., 5.)
        assert_equal(b.layer_thickness, 1.)
        assert_equal(b.cut_off_size, 5.)
        assert_equal(b.times, [0., 10000., 20000., 30000., 40000., 50000.])

    def test_set_run(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                            datfile=os.path.join(self.path, 'test3.dat'))
        a.read_pdb()
        a.read_lgt()
        a.read_dat()
        b = job.Job(a, 1., 5.)
        self.path = os.path.dirname(os.path.abspath(__file__))
        a2 = readwrite.Files(os.path.join(self.path, 'test2.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'),
                            datfile=os.path.join(self.path, 'test3.dat'))
        a2.read_pdb()
        a2.read_lgt()
        a2.read_dat()
        b.set_run(files=a2, layer_thickness=2., cut_off_size=3.)
        assert_equal(b.layer_thickness, 2.)
        assert_equal(b.cut_off_size, 3.)
        assert_equal(b.times, [0., 10000., 20000., 30000., 40000.])

    def test_set_lgts(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'))
        a.read_pdb()
        a.read_lgt()
        b = job.Job(a, 1., 5.)
        b.set_lgts()
        assert_equal(b.new_file, False)

    def test_set_times(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        a = readwrite.Files(os.path.join(self.path, 'test.pdb'), lgtfile=os.path.join(self.path, 'test.lgt'))
        a.read_pdb()
        a.read_lgt()
        b = job.Job(a, 1., 5.)
        b.set_times([0., 20000., 10000.])
        assert_equal(len(b.times), 3)
        assert_equal(b.times, [0., 10000., 20000.])

def test_check_array_true():
    array = [0, 1, 2, 3, 4]
    check = 1
    bool_ret = job.check_array(array, check)
    assert_equal(bool_ret, True)

def test_check_array_false():
    array = [0, 1, 2, 3, 4]
    check = 6
    bool_ret = job.check_array(array, check)
    assert_equal(bool_ret, False)



from falass import dataformat, job
import numpy as np
import matplotlib.pyplot as plt


class SLD:
    """SLD profile calculation.

    This class enables the calculation of the SLD profile for each of the timesteps as defined in the
    falass.job.Job. Further it will then allow the calculation and plotting of the average SLD profile.

    Parameters
    ----------
    assigned_job: falass.job.Job
        The is the Job class for the particular falass run taking place. See the job.Job class for more information.
    """
    def __init__(self, assigned_job):
        self.assigned_job = assigned_job
        self.sld_profile = []
        self.av_sld_profile = []
        self.av_sld_profile_err = []

    def get_sld_profile(self):
        """Calculate SLD profile.

        This will calculate the SLD profile for each of the timesteps defined in the falass.job.Job. This is achieved
        by summing the scattering lengths for each of the atoms found in a given layer (of defined thickness). This
        total scattering length is converted to a density by division by the volume of the layer.
        """
        prog = 0
        self.sld_profile = []
        print("[ 0 % ]")
        for i in range(0, len(self.assigned_job.files.atoms)):
            build_sld = []
            if job.check_array(self.assigned_job.times, self.assigned_job.files.times[i]):
                z_cut = self.assigned_job.files.cell[i][2] - self.assigned_job.cut_off_size
                number_of_bins = int(z_cut / self.assigned_job.layer_thickness)
                for j in range(0, number_of_bins):
                    build_sld.append(dataformat.SLDPro(self.assigned_job.layer_thickness, 0, 0))
                self.sld_profile.append(build_sld)
        k = 0
        for i in range(0, len(self.assigned_job.files.atoms)):
            if job.check_array(self.assigned_job.times, self.assigned_job.files.times[i]):
                z_cut = self.assigned_job.files.cell[i][2] - self.assigned_job.cut_off_size
                number_of_bins = int(z_cut / self.assigned_job.layer_thickness)
                for j in range(0, len(self.assigned_job.files.atoms[i])):
                    if self.assigned_job.files.atoms[i][j].zpos < number_of_bins * self.assigned_job.layer_thickness:
                        bin_choose = int(self.assigned_job.files.atoms[i][j].zpos / self.assigned_job.layer_thickness)
                        scatlen_to_add = get_scatlen(self.assigned_job.files.atoms[i][j].atom,
                                                     self.assigned_job.files.scat_lens)
                        self.sld_profile[i][bin_choose].real += scatlen_to_add[0]
                        self.sld_profile[i][bin_choose].imag += scatlen_to_add[1]
                    k += 1
                    prog_new = np.floor(k / (len(self.assigned_job.files.atoms[i]) * 
                                             len(self.assigned_job.files.atoms)) * 100)
                    if prog_new > prog + 9:
                        prog = prog_new
                        print("[{} {} % ]".format('#' * int(prog / 10), int(prog / 10) * 10))
        for i in range(0, len(self.assigned_job.files.atoms)):
            if job.check_array(self.assigned_job.times, self.assigned_job.files.times[i]) is True:
                z_cut = self.assigned_job.files.cell[i][2] - self.assigned_job.cut_off_size
                number_of_bins = int(z_cut / self.assigned_job.layer_thickness)
                for j in range(0, number_of_bins):
                    self.sld_profile[i][j].real = self.sld_profile[i][j].real / (self.assigned_job.files.cell[i][0] *
                                                                                 self.assigned_job.files.cell[i][1] *
                                                                                 self.assigned_job.layer_thickness)
                    self.sld_profile[i][j].imag = self.sld_profile[i][j].imag / (self.assigned_job.files.cell[i][0] *
                                                                                 self.assigned_job.files.cell[i][1] *
                                                                                 self.assigned_job.layer_thickness)

    def average_sld_profile(self):
        """Average SLD profiles.

        Allows for the calculation of the average SLD profile across all of the timesteps that were studied.
        """
        prog = 0
        self.av_sld_profile_err = []
        self.av_sld_profile = []
        print("[ 0 % ]")
        z_cut = self.assigned_job.files.cell[0][2] - self.assigned_job.cut_off_size
        number_of_bins = int(z_cut / self.assigned_job.layer_thickness)
        k = 0
        for j in range(0, number_of_bins):
            self.av_sld_profile.append(dataformat.SLDPro(self.assigned_job.layer_thickness, 0, 0))
            for i in range(0, len(self.assigned_job.times)):
                self.av_sld_profile[j].real += self.sld_profile[i][j].real
                self.av_sld_profile[j].imag += self.sld_profile[i][j].imag
            self.av_sld_profile[j].real /= len(self.assigned_job.times)
            self.av_sld_profile[j].imag /= len(self.assigned_job.times)
            self.av_sld_profile_err.append(dataformat.SLDPro(self.assigned_job.layer_thickness, 0, 0))
            for i in range(0, len(self.assigned_job.times)):
                self.av_sld_profile_err[j].real += np.square(self.sld_profile[i][j].real - self.av_sld_profile[j].real)
                self.av_sld_profile_err[j].imag += np.square(self.sld_profile[i][j].imag - self.av_sld_profile[j].imag)
                k += 1
                prog_new = np.floor(k / (number_of_bins * len(self.assigned_job.files.atoms)) * 100)
                if prog_new > prog + 9:
                    prog = prog_new
                    print("[{} {} % ]".format('#' * int(prog / 10), int(prog / 10) * 10))
            self.av_sld_profile_err[j].real = np.sqrt(1. / (len(self.assigned_job.times) - 1)) * self.av_sld_profile_err[j].real
            self.av_sld_profile_err[j].imag = np.sqrt(1. / (len(self.assigned_job.times) - 1)) * self.av_sld_profile_err[j].imag

    def plot_sld_profile(self, real=True, imag=False):
        """Plot SLD.

        Plots the average sld profile using matplotlib.

        Parameters
        ----------
        real: bool
            Should the real SLD profile be plotted (if both real and imaginary are true the real will be plotted).
        imag: bool
            Should the imaginary SLD profile be plotted (if both real and imaginary are true the real will be plotted).
        """
        x = []
        y = []
        dy = []
        buildx = 0
        plt.rc('text', usetex=True)
        plt.rc('font', family='serif')
        for i in range(0, len(self.av_sld_profile)):
            if real:
                y.append(self.av_sld_profile[i].real)
                dy.append(self.av_sld_profile_err[i].real)
            if imag:
                y.append(self.av_sld_profile[i].imag)
                dy.append(self.av_sld_profile_err[i].imag)
            buildx += self.av_sld_profile[i].thick
            x.append(buildx)
        plt.bar(np.asarray(x) - self.assigned_job.layer_thickness/2., np.asarray(y)*1e6, 
                width=self.assigned_job.layer_thickness, yerr=np.asarray(dy)*1e6, color='w', edgecolor='k')
        plt.ylabel('SLD (10$^{-6}$ \AA$^{-2}$)')
        plt.xlabel('$z$ (\AA)')
        plt.xlim([0, np.amax(x)])
        plt.ylim([np.amin(np.asarray(y)*1e6), np.amax(np.asarray(y)*1e6)+1])
        plt.show()


def get_scatlen(atom, scat_lens):
    """Find scattering length.

    This gets the scattering length for a given atom type

    Parameters
    ----------
    atom: str
        The name of the atom type that the scattering length is needed for.
    scat_lens: array_like falass.dataformat.ScatLens
        The array of the scattering lengths that is defined in the falass.readwrite.Files class.

    Returns
    -------
    tuple_like
        The real and imaginary scattering lengths for the given atom type.
    """
    for i in range(0, len(scat_lens)):
        if atom == scat_lens[i].atom:
            return scat_lens[i].real, scat_lens[i].imag
    raise ValueError("Attempt to get the scattering length of the atom type {} failed. This should never happen. "
                     "Please contact the developers".format(atom))



from distutils.core import setup
from setuptools import find_packages

# versioning
MAJOR = 0
MINOR = 0 
MAINT = 2
ISRELEASED = True 
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MAINT)

PACKAGES = find_packages()

# falass setup
setup(
    name='falass',
    version=VERSION,
    description='Neutron and X-ray Reflectometry from Computer Simulation',
    author='Andrew R. McCluskey',
    author_email='arm61@bath.ac.uk',
    long_description=open('README.txt').read(),
    license='MIT',
    url='http://people.bath.ac.uk/arm61/falass/',
    packages=PACKAGES,
    platforms=['Windows', 'Linux', 'Solaris', 'Mac OS-X', 'Unix'],
    install_requires=[
        'numpy', 'matplotlib', 'scipy'
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Programming Language :: Python :: 3',
    ]
)





import numpy as np
from falass import dataformat


class Files:
    """File parsing.

    Definition of the files to be used in the analysis; such as the .pdb file, .lgt file (is one exists), .dat file
    (is one exists). Also defined at this time are the resolution (this is ignored if the .dat file has 4 columns)
    and the percentage error in the intensity (this is also ignored if the .dat file has >2 columns), and whether
    the simulation cell should be flipped in the xy-plane.

    Parameters
    ----------
    pdbfile: str
        Path and name of the .pdb file to be analysed.
    lgtfile: str, optional
        Path and name of the .lgt file (which contains the scattering lengths of each of the atom types in the
        pdbfile). If a lgtfile is not defined falass will help in the creation of one.
        Currently the .lgt file style that is supported is a 3 column space separated txt file where the columns are
        atom_type, real_scattering_length, and imaginary_scattering_length respectively.
    datfile: str, optional
        Path and name of the .dat file (from which the analysis q vectors are drawn and also for subsequent
        comparison between theory and experiment). If a datfile is not defined falass will allow the user to
        define a range of q vectors to calculate the reflectometry over.
        Currently the .dat file style that is supported is a 2, 3, and 4 column space separated txt files where the
        columns are q, i, di, and dq respectively.
    resolution: float, optional
        Percentage of the q vector for the width of the resolution Gaussian function to be used in the data
        smearing, if a 4 column datfile is given this is ignored.
    ierror: float, optional
        Percentage error of the intensity to be assumed, if a >2 column datfile is used this is ignored.
    flip: bool, optional
        False if the system should be read as is, true is the simulation cell should be rotated through the
        xy-plane -- note that falass treats the first side that the neutron or X-ray interacts with as that at z=0.
    """
    def __init__(self, pdbfile, lgtfile=None, datfile=None, resolution=5., ierror=5., flip=False):
        self.pdbfile = pdbfile
        self.cell = []
        self.atoms = []
        self.number_of_timesteps = 0
        self.times = []
        self.lgtfile = lgtfile
        self.scat_lens = []
        self.datfile = datfile
        self.expdata = []
        self.ierror = ierror
        self.resolution = resolution
        self.flip = flip
        return

    def set_file(self, pdbfile=None, lgtfile=None, datfile=None):
        """Edits files.

        Let the subsequent definition, or redefinition of the pdbfile, lgtfile or datfile.

        Parameters
        ----------
        pdbfile: str
            Path and name of the .pdb file to be analysed.
        lgtfile: str, optional
            Path and name of the .lgt file (which contains the scattering lengths of each of the atom types in the
            pdbfile). If a lgtfile is not defined falass will help in the creation of one.
            Currently the .lgt file style that is supported is a 3 column space separated txt file where the columns are
            atom_type, real_scattering_length, and imaginary_scattering_length respectively.
        datfile: str, optional
            Path and name of the .dat file (from which the analysis q vectors are drawn and also for subsequent
            comparison between theory and experiment). If a datfile is not defined falass will allow the user to
            define a range of q vectors to calculate the reflectometry over.
            Currently the .dat file style that is supported is a 2, 3, and 4 column space separated txt files where the
            columns are q, i, di, and dq respectively.
        """
        if pdbfile:
            self.pdbfile = pdbfile
        if lgtfile:
            self.lgtfile = lgtfile
        if datfile:
            self.datfile = datfile
        return

    def read_pdb(self):
        """Parse .pdb.

        Reads the .pdb file into memory. Currently the atoms must have the title 'ATOM', the timestep time needs to
        be the last text in the 'TITLE' line, and the cell dimensions are taken from the 'CRYST1' line, and assumed to
        be orthorhomic. Non-orthorhomic cells are not necessarily supported.
        """
        lines = line_count(self.pdbfile)
        print("Reading PDB file \n[ 0 % ]")
        file = open(self.pdbfile, 'r')
        string = 0
        atoms_each_timestep = []
        for i, line in enumerate(file):
            string_new = np.floor(i / lines * 100)
            if string_new > string + 9:
                string = string_new
                print("[{} {} % ]".format('#' * int(string / 10), int(string/10)*10))
            if line[0:6] == "ATOM  ":
                if self.flip:
                    a = np.sqrt(np.square(self.cell[self.number_of_timesteps - 1][2] - float(line[46:54])))
                    atoms_each_timestep.append(dataformat.AtomPositions(line[12:16].strip(), a))
                else:
                    atoms_each_timestep.append(dataformat.AtomPositions(line[12:16].strip(), float(line[46:54])))
            if "TITLE  " in line:
                if self.number_of_timesteps == 0:
                    self.number_of_timesteps += 1
                    self.times.append(float(line.split()[-1]))
                else:
                    self.atoms.append(atoms_each_timestep)
                    atoms_each_timestep = []
                    self.number_of_timesteps += 1
                    self.times.append(float(line.split()[-1]))
            if "CRYST1  " in line:
                self.cell.append([float(line[6:15]), float(line[15:24]), float(line[24:33])])
        self.atoms.append(atoms_each_timestep)
        print("[{} {} % ]".format('#' * int(100 / 10), int(100)))
        file.close()
        return

    def read_lgt(self):
        """Parses .lgt.

        Parses the lgtfile. If no lgtfile is defined falass will help the user to build one by working through the
        atom types in the pdb file and requesting input of the real and imaginary scattering lengths. This will also
        occur if a atom type if found in the pdbfile but not in the given lgts file. falass will write the lgtfile 
        to disk if atom types do not feature in the given lgtfile or one is written from scratch. 
        """
        if self.lgtfile:
            lines = line_count(self.lgtfile)
            print("Reading LGT file \n[ 0 % ]")
            file = open(self.lgtfile, 'r')
            string = 0
            for i, line in enumerate(file):
                string_new = np.floor(i / lines * 100)
                if string_new > string + 9:
                    string = string_new
                    print("[{} {} % ]".format('#' * int(string / 10), int(string/10)*10))
                line_list = line.split()
                duplicate = check_duplicates(self.scat_lens, line_list[0])
                if not duplicate:
                    self.scat_lens.append(dataformat.ScatLens(line_list[0], float(line_list[1]), float(line_list[2])))
            print("[{} {} %]".format('#' * int(100 / 10), int(100)))
            file.close()
        else:
            raise ValueError("No lgtfile has been defined.")
        return

    def read_dat(self):
        """Parses .dat.

        Parses the .dat file, supporting 2, 3, and 4 column files consisting of q, i, di, and dq with comments in lines
        where the first character is a '#'. If there is no .dat file the get_qs() function should be used to generate
        q vectors to allow for the calculation of the reflectometry profile.
        """
        if self.datfile:
            lines = line_count(self.datfile)
            print("Reading DAT file \n[ 0 % ]")
            file = open(self.datfile, 'r')
            string = 0
            for i, line in enumerate(file):
                string_new = np.floor(i / lines * 100)
                if string_new > string + 9:
                    string = string_new
                    print("[{} {} % ]".format('#' * int(string / 10), int(string/10)*10))
                if line[0] != '#':
                    line_list = line.split()
                    if len(line_list) == 2:
                        self.expdata.append(dataformat.QData(float(line_list[0]), float(line_list[1]),
                                                             float(line_list[1]) * (self.ierror / 100),
                                                             float(line_list[0]) * (self.resolution / 100)))
                    if len(line_list) == 3:
                        self.expdata.append(dataformat.QData(float(line_list[0]), float(line_list[1]),
                                                             float(line_list[2]),
                                                             float(line_list[0]) * (self.resolution / 100)))
                    if len(line_list) == 4:
                        self.expdata.append(
                            dataformat.QData(float(line_list[0]), float(line_list[1]), float(line_list[2]),
                                             float(line_list[3])))
            print("[{} {} %]".format('#' * int(100 / 10), int(100)))
            file.close()
        else:
            print('No DAT file has been given, therefore no comparison will be conducted, please use the get_qs '
                  'function. Alternatively the DAT file can be added using the setFile function.')
        return

    def get_qs(self, start=0.005, end=0.5, number=50):
        """Make custom q-vectors.

        If no datfile exists this function should be used to generate a linear-spaced range of q-vector for the
        calculation of the reflectometry over.

        Parameters
        ----------
        start: float, optional
            The first q-vector for which the reflectometry should be calculated.
        end: float, optional
            The last q-vector for which the reflectometry should be calculated.
        number: int, optional
            The number of q-vectors.
        """
        q_values = np.linspace(start, end, number)
        for i in range(0, len(q_values)):
            self.expdata.append(dataformat.QData(q_values[i], None, None, q_values[i] * (self.resolution / 100)))
        return


def check_duplicates(array, check):
    """Stops duplicate atom types.

    Checks if an atom type has already been added to an array.

    Parameters
    ----------
    array: array-type ScatLens
        The array to check.
    check: str
        The atom type to try and find.

    Returns
    -------
    bool
        True if the atom type is already present in the scatlen type array, false if not.
    """
    for i in range(0, len(array)):
        if array[i].atom == check:
            return True
    return False


def line_count(filename):
    """File length.

    Quickly counts the number of lines in a file

    Parameters
    ----------
    filename: str
        Name of the file that the number of lines is desired for.

    Returns
    -------
    int
        Number of lines in the file
    """
    i = 0
    with open(filename) as f:
        for i, l in enumerate(f):
            pass
    return i + 1
